#ifndef GTREE_H 
#define GTREE_H

#include <vector>
#include <iostream>
#include <algorithm>

using namespace std;

template <typename E>
class bsTree;

template <typename E>
class bTNode{
  private:
	bTNode(E data, bTNode<E>* parent):data_(data),parent_(parent),left_(NULL),right_(NULL){}
	
	void setData(E data){data_ = data;}

	void setRightChild(E data){
		cout << "adding " << data << " as the right child of " << data_ << endl;
		if (right_!=NULL)
			right_->setData(data);
		else
			right_ = new bTNode<E>(data, this);
	}

	void setLeftChild(E data){
		cout << "adding " << data << " as the left child of " << data_ << endl;
                if (left_!=NULL)
                        left_->setData(data);
                else
                        left_ = new bTNode<E>(data, this);
                
        }

	~bTNode(){
		delete left_;
		delete right_;
	}

	friend class bsTree<E>;

	E data_;
	bTNode<E>* parent_;
	bTNode<E>* left_;
	bTNode<E>* right_;
};

template <typename E>
class bsTree{
  public:
	bsTree():root_(NULL){}

	void add(E itm){
		if (root_==NULL){
                        root_ = new bTNode<E>(itm, NULL);
			cout << "adding " << itm << " at the root!\n";
		}
		else add(itm, root_);
	}

	bool remove(E itm){
		bTNode<E>* p = findItem(itm, root_);
		if (p==NULL)
                        return false;
		return remove(p);
	}

	// assumes p is not NULL
	bool remove(bTNode<E>* p){
		if (p->left_!=NULL){
			bTNode<E>* q = findRightMostNode(p->left_);
			p->data_ = q->data_;
			return remove(q);		
		}
		
		if (p->right_!=NULL){
                        bTNode<E>* q = findLeftMostNode(p->right_);
                        p->data_ = q->data_;
                        return remove(q);
                }

		// p is a leaf
		if (p->parent_==NULL)
			root_=NULL;
		else if (p->parent_->left_==p) 
			p->parent_->left_=NULL;
		else
			p->parent_->right_= NULL;
		delete p;
		return true;
	}

	bool isEmpty() const {return root_==NULL;}

	bTNode<E>* findItem(E itm) const {return findItem(itm, root_);}

	void preorderPrint() const {preorderPrint(root_);}

	void postorderPrint() const {postorderPrint(root_);}

	void inorderPrint() const {inorderPrint(root_);}

	int size() const {return size(root_);}

	~bsTree(){ delete root_; }

  private:
        bTNode<E>* root_;

	bTNode<E>* findItem(E itm, bTNode<E>* r) const{
		if (r==NULL)
			return NULL;
        	if (r->data_==itm)
                	return r;
		else if (r->data_>itm)
			return findItem(itm, r->left_);
		else 
			return findItem(itm, r->right_);
	}

	bTNode<E>* findLeftMostNode(bTNode<E>* r) const{
		if (r->left_==NULL)
			return r;
		return findLeftMostNode(r->left_);
	}

	bTNode<E>* findRightMostNode(bTNode<E>* r) const{
                if (r->right_==NULL)
                        return r;
                return findRightMostNode(r->right_);
        }

	void preorderPrint(bTNode<E>* r) const{
     		if (r==NULL)
			return;
		cout << r->data_ << " "; // print element
		preorderPrint(r->left_);
		preorderPrint(r->right_);
	}

	void postorderPrint(bTNode<E>* r) const{
                if (r==NULL)
                        return;
                postorderPrint(r->left_);
                postorderPrint(r->right_);
        	cout << r->data_ << " "; // print element
	}

	void inorderPrint(bTNode<E>* r) const{
                if (r==NULL)
                        return;
                inorderPrint(r->left_);
		cout << r->data_ << " "; // print element
                inorderPrint(r->right_);
        }

	int size(bTNode<E>* r) const{
  		if (r==NULL)
			return 0;
  		int s = 1;
  		s+= size(r->left_);
		s+= size(r->right_);
  		return s;
	}

	// assumes r is not null
	void add(E itm, bTNode<E>* r){
		if (itm < r->data_)
			if (r->left_==NULL)
				r->setLeftChild(itm);
			else
				add(itm, r->left_);
		else 
			if (r->right_ == NULL)
				r->setRightChild(itm);
			else
				add(itm, r->right_);
	}

};

#endif


